from .rsa import (
    RSAJWKPrivateKey, RSAJWKPublicKey, generate_rsa, load_private_key,
    load_public_key, rsa_to_jwk,
)
from .token import JWT, JWTDecoder, JWTSigner
from .types import RSAPrivateKey, RSAPublicKey


__all__ = (
    "JWT",
    "JWTDecoder",
    "JWTSigner",
    "RSAJWKPrivateKey",
    "RSAJWKPublicKey",
    "RSAPrivateKey",
    "RSAPublicKey",
    "generate_rsa",
    "load_private_key",
    "load_public_key",
    "rsa_to_jwk",
)



import argparse
import logging
import os
from argparse import ArgumentParser
from pathlib import Path

from jwt_rsa.issue import parse_interval
from . import convert, issue, key_tester, keygen, pubkey, verify, jwks
from .token import ALGORITHMS


class Formatter(argparse.ArgumentDefaultsHelpFormatter, argparse.RawDescriptionHelpFormatter):
    pass


parser = ArgumentParser(formatter_class=Formatter)
parser.add_argument(
    "-a", "--algorithm", choices=ALGORITHMS,
    help="Algorithm for JWT keys", default="RS512"
)
parser.add_argument(
    "--log-level", choices=["debug", "info", "warning", "error", "critical"],
    type=lambda x: getattr(logging, x.upper(), logging.INFO), default=logging.INFO,
)

subparsers = parser.add_subparsers(dest="command", required=True)

keygen_parser = subparsers.add_parser("keygen", help="Generate a new RSA key pair", formatter_class=Formatter)
keygen_parser.set_defaults(func=keygen.main)
keygen_parser.add_argument(
    "-b", "--bits", dest="bits", type=int, default=2048, choices=[2 ** i for i in range(10, 14)],
)
keygen_parser.add_argument(
    "--kid", dest="kid", type=str, default="", help="Key ID, will be generated if missing",
)
keygen_parser.add_argument("-u", "--use", dest="use", type=str, default="sig", choices=["sig", "enc"])
keygen_parser.add_argument("-o", "--format", choices=["pem", "jwk", "base64"], default="jwk")
keygen_parser.add_argument("-r", "--raw", action="store_false", help="Print JSON with indent", dest="pretty")
keygen_parser.add_argument("-k", "--save-public", type=Path)
keygen_parser.add_argument("-K", "--save-private", type=Path)
keygen_parser.add_argument("-f", "--force", action="store_true", help="Overwrite existing keys")

keys_test_parser = subparsers.add_parser("testkey", help="Test a JWT keypair", formatter_class=Formatter)
keys_test_parser.set_defaults(func=key_tester.main)
keys_test_parser.add_argument("-K", "--private-key", type=Path, help="Private key", dest="private", required=True)
keys_test_parser.add_argument("-k", "--public-key", type=Path, help="Public key", dest="public", required=True)

pubkey_parser = subparsers.add_parser("pubkey", help="Extract public key from private key", formatter_class=Formatter)
pubkey_parser.set_defaults(func=pubkey.main)
pubkey_parser.add_argument("-K", "--private-key", type=Path, help="Private key", required=True)
pubkey_parser.add_argument("-o", "--format", choices=["pem", "jwk", "base64"], default="jwk")
pubkey_parser.add_argument("-r", "--raw", action="store_false", help="Print JSON with indent", dest="pretty")

ISSUE_PARSER_DESCRIPTION = (
    "Issue a new JWT token with claims.\n\n"
    "For time intervals you can use the following format:\n"
    "600 and +10m is means 10 minutes, +30d, -1h, +1M, -2w, etc.\n"
    "The time interval can be prefixed with '+' or '-' to indicate future or past time.\n"
    "Values without any prefix and sign are interpreted as count of seconds after current time.\n"
    "You can also use a simple integer value, which will be interpreted as seconds.\n"
    "\nThe following suffixes are supported:\n"
    " * 's' means seconds\n"
    " * 'm' means minutes\n"
    " * 'h' means hours\n"
    " * 'd' means days\n"
    " * 'w' means weeks\n"
    " * 'M' means months\n"
    " * 'y' means years\n"
)
issue_parser = subparsers.add_parser(
    "issue",
    help="Issue a new JWT token\n",
    description=ISSUE_PARSER_DESCRIPTION,
    formatter_class=Formatter
)
issue_parser.add_argument(
    "-K", "--private-key", required=True,
    help="Private JWT key", type=Path,
)
issue_parser.add_argument("--expired", help="Token expiration", type=parse_interval, default="+1M")
issue_parser.add_argument("--nbf", help="Token nbf claim", type=parse_interval, default="-1m")
issue_parser.add_argument(
    "-I", "--no-interactive", action="store_false", dest="interactive",
    help="No interactive mode, do not open editor for claims, just read JSON from stdin",
)
issue_parser.add_argument(
    "-e", "--editor", help="Editor to use in interactive mode",
    default=os.getenv("EDITOR", "vim"),
)
issue_parser.set_defaults(func=issue.main)


verify_parser = subparsers.add_parser("verify", help="Parse and verify JWT token", formatter_class=Formatter)
verify_parser.add_argument("-K", "--private-key", required=False, help="Private key", type=Path)
verify_parser.add_argument("-k", "--public-key", required=False, help="Public key", type=Path)
verify_parser.add_argument("-V", "--no-verify", action="store_false", help="No verify signature", dest="verify")
verify_parser.add_argument(
    "-I", "--no-interactive", action="store_false", dest="interactive",
    help="Interactive mode or raw read token from stdin",
)
verify_parser.set_defaults(func=verify.main)


convert_parser = subparsers.add_parser("convert", help="Convert JWT token", formatter_class=Formatter)
convert_parser.set_defaults(func=convert.main)
convert_parser.add_argument("private_key", help="Private key source", type=Path)
convert_parser.add_argument("-k", "--save-public", type=Path)
convert_parser.add_argument("-K", "--save-private", type=Path)
convert_parser.add_argument("-o", "--format", choices=["pem", "jwk", "base64"], default="jwk")
convert_parser.add_argument("-f", "--force", action="store_true", help="Overwrite existing keys")
convert_parser.add_argument("-r", "--raw", action="store_false", help="Print JSON with indent", dest="pretty")

jwks_parser = subparsers.add_parser("jwks", help="Fetch JWKs from remote hosts", formatter_class=Formatter)
jwks_parser.set_defaults(func=jwks.main)
jwks_parser.add_argument("url", help="URL for JWKs", type=str)


def main() -> None:
    args = parser.parse_args()
    logging.basicConfig(level=args.log_level, format="%(levelname)s: %(message)s")
    args.func(args)


if __name__ == "__main__":
    main()



import base64
import json
from pathlib import Path
from typing import NamedTuple, Optional, TypedDict, overload

from cryptography.hazmat.backends import default_backend

from .types import (
    AlgorithmType, RSAPrivateKey, RSAPublicKey, rsa, serialization,
)


class KeyPair(NamedTuple):
    private: RSAPrivateKey
    public: RSAPublicKey


class JWKKeyPair(NamedTuple):
    private: Optional[RSAPrivateKey]
    public: RSAPublicKey


class RSAJWKPublicKey(TypedDict):
    kty: str
    e: str
    n: str
    kid: str
    alg: str
    use: str


class RSAJWKPrivateKey(RSAJWKPublicKey):
    d: str
    p: str
    q: str
    dp: str
    dq: str
    qi: str


def generate_rsa(bits: int = 2048) -> KeyPair:
    """
    Generate an RSA keypair with an exponent of 65537 in PEM format
    param: bits The key length in bits
    Return private key and public key
    """
    private_key = rsa.generate_private_key(public_exponent=65537, key_size=bits, backend=default_backend())
    public_key = private_key.public_key()
    return KeyPair(private=private_key, public=public_key)


def base64url_decode(data: str) -> bytes:
    rem = len(data) % 4
    if rem > 0:
        data += "=" * (4 - rem)
    return base64.urlsafe_b64decode(data)


def load_jwk_public_key(jwk: RSAJWKPublicKey) -> RSAPublicKey:
    if jwk["kty"] != "RSA":
        raise ValueError("Not an RSA key")

    e = int.from_bytes(base64url_decode(jwk["e"]), "big")
    n = int.from_bytes(base64url_decode(jwk["n"]), "big")

    return rsa.RSAPublicNumbers(e, n).public_key(default_backend())


def load_jwk_private_key(jwk: RSAJWKPrivateKey) -> RSAPrivateKey:
    if jwk["kty"] != "RSA":
        raise ValueError("Not an RSA key")

    e = int.from_bytes(base64url_decode(jwk["e"]), "big")
    n = int.from_bytes(base64url_decode(jwk["n"]), "big")
    d = int.from_bytes(base64url_decode(jwk["d"]), "big")
    p = int.from_bytes(base64url_decode(jwk["p"]), "big")
    q = int.from_bytes(base64url_decode(jwk["q"]), "big")
    dp = int.from_bytes(base64url_decode(jwk["dp"]), "big")
    dq = int.from_bytes(base64url_decode(jwk["dq"]), "big")
    qi = int.from_bytes(base64url_decode(jwk["qi"]), "big")

    public_numbers = rsa.RSAPublicNumbers(e, n)
    private_numbers = rsa.RSAPrivateNumbers(p, q, d, dp, dq, qi, public_numbers)

    return private_numbers.private_key(default_backend())


def load_jwk(jwk: RSAJWKPublicKey | RSAJWKPrivateKey | str) -> JWKKeyPair:
    jwk_dict: RSAJWKPublicKey | RSAJWKPrivateKey

    if isinstance(jwk, str):
        jwk_dict = json.loads(jwk)
    else:
        jwk_dict = jwk

    if "d" in jwk_dict:  # Private key
        private_key = load_jwk_private_key(jwk_dict)    # type: ignore
        public_key = private_key.public_key()
    else:  # Public key
        public_key = load_jwk_public_key(jwk_dict)
        private_key = None

    return JWKKeyPair(private=private_key, public=public_key)


def int_to_base64url(value: int) -> str:
    return base64.urlsafe_b64encode(
        value.to_bytes((value.bit_length() + 7) // 8, byteorder="big"),
    ).rstrip(b"=").decode("ascii")


@overload
def rsa_to_jwk(
    key: RSAPublicKey, *, kid: str = "", alg: AlgorithmType = "RS256", use: str = "sig",
) -> RSAJWKPublicKey: ...


@overload
def rsa_to_jwk(
    key: RSAPrivateKey, *, kid: str = "", alg: AlgorithmType = "RS256", use: str = "sig",
) -> RSAJWKPrivateKey: ...


def rsa_to_jwk(
    key: RSAPrivateKey | RSAPublicKey,
    *,
    kid: str = "",
    alg: AlgorithmType = "RS256",
    use: str = "sig",
    kty: str = "RSA",
) -> RSAJWKPublicKey | RSAJWKPrivateKey:
    if isinstance(key, RSAPublicKey):
        public_numbers = key.public_numbers()
        private_numbers = None
    elif isinstance(key, RSAPrivateKey):
        public_numbers = key.public_key().public_numbers()
        private_numbers = key.private_numbers()
    else:
        raise ValueError("Invalid key type: {}".format(type(key)))

    result = RSAJWKPublicKey(
        kty=kty,
        kid=kid,
        alg=alg,
        use=use,
        n=int_to_base64url(public_numbers.n),
        e=int_to_base64url(public_numbers.e),
    )

    if private_numbers is None:
        return result

    return RSAJWKPrivateKey(
        kty=kty,
        kid=kid,
        alg=alg,
        use=use,
        n=int_to_base64url(public_numbers.n),
        e=int_to_base64url(public_numbers.e),
        d=int_to_base64url(private_numbers.d),
        p=int_to_base64url(private_numbers.p),
        q=int_to_base64url(private_numbers.q),
        dp=int_to_base64url(private_numbers.dmp1),
        dq=int_to_base64url(private_numbers.dmq1),
        qi=int_to_base64url(private_numbers.iqmp),
    )


def load_private_key(data: str | RSAJWKPrivateKey | Path) -> RSAPrivateKey:
    if isinstance(data, Path):
        data = data.read_text()
    if isinstance(data, str):
        if data.startswith("-----BEGIN "):
            result = serialization.load_pem_private_key(data.encode(), None, default_backend())
            assert isinstance(result, RSAPrivateKey)
            return result
        if data.strip().startswith("{"):
            return load_jwk_private_key(json.loads(data))
    if isinstance(data, dict):
        return load_jwk_private_key(json.loads(json.dumps(data)))
    key = serialization.load_der_private_key(base64.b64decode(data), None, default_backend())
    if not isinstance(key, RSAPrivateKey):
        raise ValueError("Key {!r} is not an RSA key".format(key))
    return key


def load_public_key(data: str | RSAJWKPublicKey | Path) -> RSAPublicKey:
    if isinstance(data, Path):
        data = data.read_text()
    if isinstance(data, str):
        if data.startswith("-----BEGIN "):
            result = serialization.load_pem_public_key(data.encode(), default_backend())
            assert isinstance(result, RSAPublicKey)
            return result
        if data.strip().startswith("{"):
            return load_jwk_public_key(json.loads(data))
    if isinstance(data, dict):
        return load_jwk_public_key(json.loads(json.dumps(data)))
    key = serialization.load_der_public_key(base64.b64decode(data), default_backend())
    if not isinstance(key, RSAPublicKey):
        raise ValueError("Key {!r} is not an RSA key".format(key))
    return key


__all__ = (
    "RSAJWKPrivateKey",
    "RSAJWKPublicKey",
    "KeyPair",
    "base64url_decode",
    "generate_rsa",
    "int_to_base64url",
    "load_jwk",
    "load_jwk_private_key",
    "load_jwk_public_key",
    "load_private_key",
    "load_public_key",
    "rsa_to_jwk",
)



from datetime import datetime, timedelta
from typing import Literal

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPrivateKeyWithSerialization as RSAPrivateKey,
)
from cryptography.hazmat.primitives.asymmetric.rsa import (
    RSAPublicKeyWithSerialization as RSAPublicKey,
)


AlgorithmType = Literal["RS256", "RS384", "RS512"]
DateType = timedelta | datetime | float | int

__all__ = (
    "AlgorithmType",
    "DateType",
    "RSAPrivateKey",
    "RSAPublicKey",
    "serialization",
    "rsa",
)



import argparse
import json
import os
import platform
import pwd
import re
import sys
import time
from datetime import datetime
from subprocess import check_call
from tempfile import NamedTemporaryFile
from types import SimpleNamespace

from .rsa import load_private_key
from .token import JWT


TEMPLATE = """# THIS FILE SUPPORTS COMMENTS AND TRAILING COMMAS
# Actually it's a Python dictionary that will be evaluated as a JSON object
# Anyway it supports Python-style single and double quotes, math expressions, etc.

%(preamble)s
{
    # === Standard JWT Claims (As per RFC 7519) ===

    # Issuer of the token
    "iss": "{} <{}@{}>".format(whoami.pw_name, whoami.pw_gecos, HOSTNAME),

    # Subject of the token (usually the user ID)
    "sub": "JohnDoe <johndoe@localhost>",

    # Audience for the token
    "aud": "your-audience",

    # Expiration time (Unix timestamp)
    "exp": timestamp + %(exp)d,

    # Not before time (Unix timestamp)
    # Best practice is to set it to current time minus 1 minute to avoid clock skew
    # "nbf": timestamp + (%(nbf)d),

    # Issued at time (Unix timestamp)
    # "iat": int(time.time()),

    # JWT ID - Unique identifier for the token
    # "jti": "unique-token-id",

    # ===      Custom/User-Defined Claims       ===
    # === Not a part of the standard JWT claims ===

    # Username of the authenticated user
    "username": "johndoe",

    # Roles assigned to the user
    "roles": ["admin", "user"],

    # Optional custom fields (commented out)

    # "email": "johndoe@example.com",
    # "permissions": ["read", "write", "delete"],
    # "profile": {
    #   "firstName": "John",
    #   "lastName": "Doe",
    #   "age": 30
    # },
    # "lastLogin": "2025-01-02T12:34:56Z",
    # "theme": "dark",
    # "locale": "en-US"
}
"""


TIMEINTERVAL_EXP = re.compile(r"^((?P<is_interval>[+-])?(?P<value>[0-9]+)(?P<suffix>[smhdMy]?))+$")
SUFFIXES = {"s": 1, "m": 60, "h": 3600, "d": 86400, "M": 2592000, "y": 31536000}


def parse_interval(value: str) -> int:
    match = TIMEINTERVAL_EXP.match(value)
    if not match:
        raise argparse.ArgumentTypeError(f"Invalid time interval format: {value}")

    groups = match.groupdict()
    if groups["is_interval"] is not None:
        seconds = int(groups["value"])
        if groups["is_interval"] == "-":
            seconds = -seconds
        suffix = SUFFIXES.get(groups["suffix"], 1)
        return int(seconds * suffix)
    elif groups["is_interval"] is None and groups["suffix"]:
        raise argparse.ArgumentTypeError("Invalid time interval format, missing sign: " + value)
    else:
        return int(int(groups["value"]))


def main(arguments: SimpleNamespace) -> None:
    jwt = JWT(
        load_private_key(arguments.private_key),
        expires=arguments.expired,
        nbf_delta=-arguments.nbf,
        algorithm=arguments.algorithm
    )

    whoami = pwd.getpwuid(os.getuid())

    if arguments.interactive:
        globals = {
            "HOUR": 3600,
            "DAY": 86400,
            "YEAR": 365 * 86400,
            "HOSTNAME": platform.node(),
            "datetime": datetime,
            "now": datetime.now(),
            "timestamp": int(time.time()),
            "time": time,
            "int": int,
            "format": format,
            "sum": sum,
            "whoami": whoami,
        }

        preable = "# This modules functions and constants are available:\n\n"
        for key, value in sorted(globals.items(), key=lambda x: x[0]):
            preable += f"#  * {key} = {value!r}\n"

        with NamedTemporaryFile("wt", suffix=".py") as fp:
            if arguments.interactive:
                fp.write(
                    TEMPLATE % {
                        "exp": arguments.expired,
                        "nbf": arguments.nbf,
                        "preamble": preable,
                    },
                )
                fp.flush()

                while True:
                    check_call([arguments.editor, fp.name])

                    try:
                        claims = eval(open(fp.name).read(), globals, {})
                        break
                    except (ValueError, TypeError) as exc:
                        print(f"Error parsing JSON: {exc}", file=sys.stderr)
                        input("Press Enter to try again... Press Ctrl+C to abort")
            else:
                claims = json.load(sys.stdin)
    else:
        claims = json.loads(sys.stdin.read())
    print(jwt.encode(**claims))



import argparse
import http.client
import json
import logging
import ssl
from abc import ABC, abstractmethod
from contextlib import closing
from functools import lru_cache
from typing import TypedDict, Any
from urllib.parse import urlparse

from jwt.api_jws import PyJWS

from . import JWTDecoder
from .rsa import RSAJWKPublicKey, load_jwk_public_key
from .types import RSAPublicKey

log = logging.getLogger(__name__)


class JWKsStructure(TypedDict, total=True):
    keys: list[RSAJWKPublicKey]


class JWKFetcher(ABC):
    keys: dict[str, RSAPublicKey]
    ca: dict[str, RSAPublicKey]

    def __init__(self, url: str, **kwargs: Any) -> None:
        self.url = url
        self.keys = {}
        self.ca = {}
        self.jws = PyJWS(**kwargs)

    def load_key(self, key: RSAJWKPublicKey) -> RSAPublicKey | None:
        return load_jwk_public_key(key)

    def parse_jwk(self, jwk: dict[str, Any]) -> dict[str, RSAPublicKey]:
        keys = {}
        for key in jwk["keys"]:
            jwk = RSAJWKPublicKey(**key)  # type: ignore
            loaded_key = self.load_key(key)
            if loaded_key is None:
                log.warning("Skipping JWK key: %s", key)
                continue
            keys[jwk['kid']] = loaded_key
        return keys

    @abstractmethod
    def refresh(self) -> Any:
        raise NotImplementedError()

    @lru_cache(1024)
    def decoder(self, kid: str) -> JWTDecoder:
        if kid not in self.keys:
            raise ValueError(f"Key with kid '{kid}' is unknown")
        return JWTDecoder(key=self.keys[kid])

    def decode(self, token: str) -> dict[str, Any]:
        header = self.jws.get_unverified_header(token)
        if 'kid' not in header:
            raise ValueError("Token does not contain 'kid' header")

        decoder = self.decoder(header['kid'])
        return decoder.decode(token)


class HTTPSJWKFetcher(JWKFetcher):
    CLIENT_CLASS = http.client.HTTPSConnection

    def __init__(self, url: str, *, ssl_context: ssl.SSLContext | None = None, **kwargs: Any) -> None:
        super().__init__(url, **kwargs)
        self.ssl_context = ssl_context or ssl.create_default_context()

    def refresh(self) -> None:
        url_parts = urlparse(self.url)
        host, port = (
            url_parts.netloc.split(":")
            if ":" in url_parts.netloc
            else (url_parts.netloc, 443)
        )
        with closing(self.CLIENT_CLASS(host, int(port), context=self.ssl_context)) as client:
            client.request("GET", url_parts.path)
            response = client.getresponse()
            if response.status != 200:
                raise ValueError(f"Failed to fetch JWKs: {response.status} {response.reason}")
            self.keys = self.parse_jwk(json.load(response))
            # Clear the cache for the decoder to ensure it uses the latest keys
            self.decoder.cache_clear()


def main(args: argparse.Namespace) -> None:
    from .rsa import rsa_to_jwk
    fetcher = HTTPSJWKFetcher(args.url)
    fetcher.refresh()

    log.info("JWKs fetched from %r found %s keys", fetcher.url, len(fetcher.keys))

    for kid, key in fetcher.keys.items():
        log.info("Key ID: %s", kid)
        print(json.dumps(rsa_to_jwk(key)))



import logging
from types import SimpleNamespace

from .convert import convert
from .rsa import load_private_key


def main(arguments: SimpleNamespace) -> None:
    logging.basicConfig(level=logging.INFO)

    private_key = load_private_key(arguments.private_key)
    public_key = private_key.public_key()

    public_data, _ = convert(
        private_key,
        public_key,
        arguments.format,
        pretty=arguments.pretty,
    )

    print(public_data)
